/*
Computer Graphics Programming in OpenGL with Java, 2nd Edition, V. Scott Gordon, John Clevenger
Pages 85-86.
*/

import static com.jogamp.opengl.GL4.*;
import com.jogamp.opengl.*;
import com.jogamp.opengl.awt.GLJPanel;
import com.jogamp.common.nio.Buffers;

import org.joml.*;

import java.nio.*;
import java.lang.Math;
import javax.swing.*;

public class Code_13 extends JFrame implements GLEventListener
{
   private final GLJPanel myCanvas;
   private final int[] vao = new int[1];
   private final int[] vbo = new int[2];
   private float cameraX, cameraY, cameraZ;
   private float cubeLocX, cubeLocY, cubeLocZ;
   private float pyrLocX, pyrLocY, pyrLocZ;
   private int renderingProgram;

   // allocate variables for display() function
   private final FloatBuffer vals = Buffers.newDirectFloatBuffer(16);
   private final Matrix4f pMat = new Matrix4f();  // perspective matrix
   private final Matrix4f vMat = new Matrix4f();  // view matrix
   private final Matrix4f mMat = new Matrix4f();  // model matrix
   private final Matrix4f mvMat = new Matrix4f(); // model-view matrix
   private int mvLoc, projLoc;
   private float aspect;

   public Code_13()
   {
      this.setTitle("Chapter 4 - program 3 (Code_13)");
      this.setSize(600, 600);
      myCanvas = new GLJPanel();
      myCanvas.addGLEventListener(this);
      this.add(myCanvas);
      this.setLocationRelativeTo(null);
      this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      this.setVisible(true);
   }

   public void display(GLAutoDrawable drawable)
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      gl.glClear(GL_COLOR_BUFFER_BIT);
      gl.glClear(GL_DEPTH_BUFFER_BIT);

      gl.glUseProgram(renderingProgram);

        mvLoc = gl.glGetUniformLocation(renderingProgram, "mv_matrix");
      projLoc = gl.glGetUniformLocation(renderingProgram, "proj_matrix");

      aspect = (float) myCanvas.getWidth() / (float) myCanvas.getHeight();
      pMat.setPerspective((float) Math.toRadians(60.0f), aspect, 0.1f, 1000.0f);

      vMat.translation(-cameraX,-cameraY,-cameraZ);

      // draw the cube using buffer #0

      mMat.translation(cubeLocX, cubeLocY, cubeLocZ);

      mvMat.identity();
      mvMat.mul(vMat);
      mvMat.mul(mMat);

      gl.glUniformMatrix4fv(  mvLoc, 1, false, mvMat.get(vals));
      gl.glUniformMatrix4fv(projLoc, 1, false, pMat.get(vals));

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
      gl.glVertexAttribPointer(0, 3, GL_FLOAT, false, 0, 0);
      gl.glEnableVertexAttribArray(0);

      gl.glEnable(GL_DEPTH_TEST);
      gl.glDepthFunc(GL_LEQUAL);

      gl.glDrawArrays(GL_TRIANGLES, 0, 36);

      // draw the pyramid using buffer #1

      mMat.translation(pyrLocX, pyrLocY, pyrLocZ);

      mvMat.identity();
      mvMat.mul(vMat);
      mvMat.mul(mMat);

      gl.glUniformMatrix4fv(  mvLoc, 1, false, mvMat.get(vals));
      gl.glUniformMatrix4fv(projLoc, 1, false,  pMat.get(vals));

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
      gl.glVertexAttribPointer(0, 3, GL_FLOAT, false, 0, 0);
      gl.glEnableVertexAttribArray(0);

      gl.glEnable(GL_DEPTH_TEST);
      gl.glDepthFunc(GL_LEQUAL);

      gl.glDrawArrays(GL_TRIANGLES, 0, 18);
   }

   public void init(GLAutoDrawable drawable)
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      renderingProgram = Utils.createShaderProgram("shaders/vertShader_9.glsl",
                                                   "shaders/fragShader_9.glsl");
      setupVertices();

       cameraX = 0.0f;  cameraY =  0.0f;  cameraZ = 8.0f;
      cubeLocX = 0.0f; cubeLocY = -2.0f; cubeLocZ = 0.0f;

      pyrLocX = 2.0f; pyrLocY = 2.0f; pyrLocZ = 0.0f;
   }

   private void setupVertices()
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      final float[] cubePositions =
      {
        -1.0f,  1.0f, -1.0f, -1.0f, -1.0f, -1.0f,  1.0f, -1.0f, -1.0f,
         1.0f, -1.0f, -1.0f,  1.0f,  1.0f, -1.0f, -1.0f,  1.0f, -1.0f,
         1.0f, -1.0f, -1.0f,  1.0f, -1.0f,  1.0f,  1.0f,  1.0f, -1.0f,
         1.0f, -1.0f,  1.0f,  1.0f,  1.0f,  1.0f,  1.0f,  1.0f, -1.0f,
         1.0f, -1.0f,  1.0f, -1.0f, -1.0f,  1.0f,  1.0f,  1.0f,  1.0f,
        -1.0f, -1.0f,  1.0f, -1.0f,  1.0f,  1.0f,  1.0f,  1.0f,  1.0f,
        -1.0f, -1.0f,  1.0f, -1.0f, -1.0f, -1.0f, -1.0f,  1.0f,  1.0f,
        -1.0f, -1.0f, -1.0f, -1.0f,  1.0f, -1.0f, -1.0f,  1.0f,  1.0f,
        -1.0f, -1.0f,  1.0f,  1.0f, -1.0f,  1.0f,  1.0f, -1.0f, -1.0f,
         1.0f, -1.0f, -1.0f, -1.0f, -1.0f, -1.0f, -1.0f, -1.0f,  1.0f,
        -1.0f,  1.0f, -1.0f,  1.0f,  1.0f, -1.0f,  1.0f,  1.0f,  1.0f,
         1.0f,  1.0f,  1.0f, -1.0f,  1.0f,  1.0f, -1.0f,  1.0f, -1.0f
      };

      final float[] pyramidPositions =
      {
        -1.0f, -1.0f,  1.0f,  1.0f, -1.0f,  1.0f,  0.0f,  1.0f,  0.0f, //front
         1.0f, -1.0f,  1.0f,  1.0f, -1.0f, -1.0f,  0.0f,  1.0f,  0.0f, //right
         1.0f, -1.0f, -1.0f, -1.0f, -1.0f, -1.0f,  0.0f,  1.0f,  0.0f, //back
        -1.0f, -1.0f, -1.0f, -1.0f, -1.0f,  1.0f,  0.0f,  1.0f,  0.0f, //left
        -1.0f, -1.0f, -1.0f,  1.0f, -1.0f,  1.0f, -1.0f, -1.0f,  1.0f, //LF
         1.0f, -1.0f,  1.0f, -1.0f, -1.0f, -1.0f,  1.0f, -1.0f, -1.0f  //RR
      };

      gl.glGenVertexArrays(vao.length, vao, 0);
      gl.glBindVertexArray(vao[0]);
      gl.glGenBuffers(vbo.length, vbo, 0);

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
      FloatBuffer cubeBuf = Buffers.newDirectFloatBuffer(cubePositions);
      gl.glBufferData(GL_ARRAY_BUFFER, cubeBuf.limit()*4, cubeBuf, GL_STATIC_DRAW);

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
      FloatBuffer pyrBuf = Buffers.newDirectFloatBuffer(pyramidPositions);
      gl.glBufferData(GL_ARRAY_BUFFER, pyrBuf.limit()*4, pyrBuf, GL_STATIC_DRAW);
   }

   public void reshape(GLAutoDrawable drawable, int x, int y, int w, int h){}
   public void dispose(GLAutoDrawable drawable){}

   public static void main(String[] args)
   {
      new Code_13();
   }
}
