/*
Computer Graphics Programming in OpenGL with Java, 2nd Edition, V. Scott Gordon, John Clevenger
Pages 92-93.
*/

import static com.jogamp.opengl.GL4.*;
import com.jogamp.opengl.*;
import com.jogamp.opengl.awt.GLJPanel;
import com.jogamp.common.nio.Buffers;
import com.jogamp.opengl.util.*;

import org.joml.*;

import java.nio.*;
import java.lang.Math;
import javax.swing.*;

public class Code_14 extends JFrame implements GLEventListener
{
   private final GLJPanel myCanvas;
   private final int[] vao = new int[1];
   private final int[] vbo = new int[2];
   private float cameraX, cameraY, cameraZ;
   private int renderingProgram;
   private double startTime = 0.0;
   private double elapsedTime;

   // allocate variables for display() function
   private final FloatBuffer vals = Buffers.newDirectFloatBuffer(16);
   private final Matrix4fStack mvStack = new Matrix4fStack(5);
   private final Matrix4f pMat = new Matrix4f();
   private int mvLoc, projLoc;
   private float aspect;
   private double tf;

   public Code_14()
   {
      this.setTitle("Chapter 4 - program 4 (Code_14)");
      this.setSize(600, 600);
      myCanvas = new GLJPanel();
      myCanvas.addGLEventListener(this);
      this.add(myCanvas);
      this.setLocationRelativeTo(null);
      this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      this.setVisible(true);

      final Animator animator = new Animator(myCanvas);
      animator.start();
   }

   public void display(GLAutoDrawable drawable)
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      gl.glClear(GL_COLOR_BUFFER_BIT);
      gl.glClear(GL_DEPTH_BUFFER_BIT);
      elapsedTime = System.currentTimeMillis() - startTime;

      gl.glUseProgram(renderingProgram);

        mvLoc = gl.glGetUniformLocation(renderingProgram, "mv_matrix");
      projLoc = gl.glGetUniformLocation(renderingProgram, "proj_matrix");

      aspect = (float) myCanvas.getWidth() / (float) myCanvas.getHeight();
      pMat.identity().setPerspective((float) Math.toRadians(60.0f), aspect, 0.1f, 1000.0f);
      gl.glUniformMatrix4fv(projLoc, 1, false, pMat.get(vals));

      // push view matrix onto the stack
      mvStack.pushMatrix();
      mvStack.translate(-cameraX, -cameraY, -cameraZ);

      tf = elapsedTime/1000.0;  // time factor

      // ----------------------  pyramid == sun
      mvStack.pushMatrix();
      mvStack.translate(0.0f, 0.0f, 0.0f);
      mvStack.pushMatrix();
      mvStack.rotate((float)tf, 1.0f, 0.0f, 0.0f);
      gl.glUniformMatrix4fv(mvLoc, 1, false, mvStack.get(vals));
      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
      gl.glVertexAttribPointer(0, 3, GL_FLOAT, false, 0, 0);
      gl.glEnableVertexAttribArray(0);
      gl.glEnable(GL_DEPTH_TEST);
      gl.glDrawArrays(GL_TRIANGLES, 0, 18);
      mvStack.popMatrix();

      //-----------------------  cube == planet
      mvStack.pushMatrix();
      mvStack.translate((float)Math.sin(tf)*4.0f, 0.0f, (float)Math.cos(tf)*4.0f);
      mvStack.pushMatrix();
      mvStack.rotate((float)tf, 0.0f, 1.0f, 0.0f);
      gl.glUniformMatrix4fv(mvLoc, 1, false, mvStack.get(vals));
      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
      gl.glVertexAttribPointer(0, 3, GL_FLOAT, false, 0, 0);
      gl.glEnableVertexAttribArray(0);
      gl.glDrawArrays(GL_TRIANGLES, 0, 36);
      mvStack.popMatrix();

      //-----------------------  smaller cube == moon
      mvStack.pushMatrix();
      mvStack.translate(0.0f, (float)Math.sin(tf)*2.0f, (float)Math.cos(tf)*2.0f);
      mvStack.rotate((float)tf, 0.0f, 0.0f, 1.0f);
      mvStack.scale(0.25f, 0.25f, 0.25f);
      gl.glUniformMatrix4fv(mvLoc, 1, false, mvStack.get(vals));
      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
      gl.glVertexAttribPointer(0, 3, GL_FLOAT, false, 0, 0);
      gl.glEnableVertexAttribArray(0);
      gl.glDrawArrays(GL_TRIANGLES, 0, 36);
      mvStack.popMatrix();
      mvStack.popMatrix();
      mvStack.popMatrix();
      mvStack.popMatrix();
   }

   public void init(GLAutoDrawable drawable)
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      startTime = System.currentTimeMillis();

      renderingProgram = Utils.createShaderProgram("shaders/vertShader_9.glsl",
                                                   "shaders/fragShader_9.glsl");
      setupVertices();

      cameraX = 0.0f; cameraY = 0.0f; cameraZ = 12.0f;
   }

   private void setupVertices()
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      final float[] cubePositions =
      {
        -1.0f,  1.0f, -1.0f, -1.0f, -1.0f, -1.0f,  1.0f, -1.0f, -1.0f,
         1.0f, -1.0f, -1.0f,  1.0f,  1.0f, -1.0f, -1.0f,  1.0f, -1.0f,
         1.0f, -1.0f, -1.0f,  1.0f, -1.0f,  1.0f,  1.0f,  1.0f, -1.0f,
         1.0f, -1.0f,  1.0f,  1.0f,  1.0f,  1.0f,  1.0f,  1.0f, -1.0f,
         1.0f, -1.0f,  1.0f, -1.0f, -1.0f,  1.0f,  1.0f,  1.0f,  1.0f,
        -1.0f, -1.0f,  1.0f, -1.0f,  1.0f,  1.0f,  1.0f,  1.0f,  1.0f,
        -1.0f, -1.0f,  1.0f, -1.0f, -1.0f, -1.0f, -1.0f,  1.0f,  1.0f,
        -1.0f, -1.0f, -1.0f, -1.0f,  1.0f, -1.0f, -1.0f,  1.0f,  1.0f,
        -1.0f, -1.0f,  1.0f,  1.0f, -1.0f,  1.0f,  1.0f, -1.0f, -1.0f,
         1.0f, -1.0f, -1.0f, -1.0f, -1.0f, -1.0f, -1.0f, -1.0f,  1.0f,
        -1.0f,  1.0f, -1.0f,  1.0f,  1.0f, -1.0f,  1.0f,  1.0f,  1.0f,
         1.0f,  1.0f,  1.0f, -1.0f,  1.0f,  1.0f, -1.0f,  1.0f, -1.0f
      };

      final float[] pyramidPositions =
      {
        -1.0f, -1.0f,  1.0f,  1.0f, -1.0f,  1.0f,  0.0f,  1.0f,  0.0f, //front
         1.0f, -1.0f,  1.0f,  1.0f, -1.0f, -1.0f,  0.0f,  1.0f,  0.0f, //right
         1.0f, -1.0f, -1.0f, -1.0f, -1.0f, -1.0f,  0.0f,  1.0f,  0.0f, //back
        -1.0f, -1.0f, -1.0f, -1.0f, -1.0f,  1.0f,  0.0f,  1.0f,  0.0f, //left
        -1.0f, -1.0f, -1.0f,  1.0f, -1.0f,  1.0f, -1.0f, -1.0f,  1.0f, //LF
         1.0f, -1.0f,  1.0f, -1.0f, -1.0f, -1.0f,  1.0f, -1.0f, -1.0f  //RR
      };

      gl.glGenVertexArrays(vao.length, vao, 0);
      gl.glBindVertexArray(vao[0]);
      gl.glGenBuffers(vbo.length, vbo, 0);

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
      FloatBuffer cubeBuf = Buffers.newDirectFloatBuffer(cubePositions);
      gl.glBufferData(GL_ARRAY_BUFFER, cubeBuf.limit()*4, cubeBuf, GL_STATIC_DRAW);

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
      FloatBuffer pyrBuf = Buffers.newDirectFloatBuffer(pyramidPositions);
      gl.glBufferData(GL_ARRAY_BUFFER, pyrBuf.limit()*4, pyrBuf, GL_STATIC_DRAW);
   }

   public void reshape(GLAutoDrawable drawable, int x, int y, int w, int h){}
   public void dispose(GLAutoDrawable drawable){}

   public static void main(String[] args)
   {
      new Code_14();
   }
}
