/*
Computer Graphics Programming in OpenGL with Java, 2nd Edition, V. Scott Gordon, John Clevenger
Pages 129-130.
*/

import static com.jogamp.opengl.GL4.*;
import com.jogamp.opengl.*;
import com.jogamp.opengl.awt.GLJPanel;
import com.jogamp.common.nio.Buffers;

import org.joml.*;

import java.io.*;
import java.nio.*;
import java.lang.Math;
import javax.swing.*;

import javax.imageio.ImageIO;
import java.awt.image.*;
import java.awt.color.ColorSpace;
import java.awt.geom.AffineTransform;
import java.awt.Graphics2D;

public class Code_16 extends JFrame implements GLEventListener
{
   private final GLJPanel myCanvas;
   private final int[] vao = new int[1];
   private final int[] vbo = new int[2];
   private float cameraX, cameraY, cameraZ;
   private float pyrLocX, pyrLocY, pyrLocZ;
   private int renderingProgram;
   private int brickTexture;

   // allocate variables for display() function
   private final FloatBuffer vals = Buffers.newDirectFloatBuffer(16);
   private final Matrix4f pMat = new Matrix4f();  // perspective matrix
   private final Matrix4f vMat = new Matrix4f();  // view matrix
   private final Matrix4f mMat = new Matrix4f();  // model matrix
   private final Matrix4f mvMat = new Matrix4f(); // model-view matrix
   private int mvLoc, projLoc;
   private float aspect;

   public Code_16()
   {
      this.setTitle("Chapter 5 - program 2 (Code_16)");
      this.setSize(800, 800);
      myCanvas = new GLJPanel();
      myCanvas.addGLEventListener(this);
      this.add(myCanvas);
      this.setLocationRelativeTo(null);
      this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      this.setVisible(true);
   }

   public void display(GLAutoDrawable drawable)
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      gl.glClear(GL_COLOR_BUFFER_BIT);
      gl.glClear(GL_DEPTH_BUFFER_BIT);

      gl.glUseProgram(renderingProgram);

        mvLoc = gl.glGetUniformLocation(renderingProgram, "mv_matrix");
      projLoc = gl.glGetUniformLocation(renderingProgram, "proj_matrix");

      vMat.identity().setTranslation(-cameraX, -cameraY, -cameraZ);

      mMat.identity();
      mMat.translate(pyrLocX, pyrLocY, pyrLocZ);
      mMat.rotateXYZ(-0.45f, 0.61f, 0.0f);

      mvMat.identity();
      mvMat.mul(vMat);
      mvMat.mul(mMat);

      gl.glUniformMatrix4fv(  mvLoc, 1, false, mvMat.get(vals));
      gl.glUniformMatrix4fv(projLoc, 1, false,  pMat.get(vals));

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
      gl.glVertexAttribPointer(0, 3, GL_FLOAT, false, 0, 0);
      gl.glEnableVertexAttribArray(0);

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
      gl.glVertexAttribPointer(1, 2, GL_FLOAT, false, 0, 0);
      gl.glEnableVertexAttribArray(1);

      gl.glActiveTexture(GL_TEXTURE0);
      gl.glBindTexture(GL_TEXTURE_2D, brickTexture);

      gl.glEnable(GL_DEPTH_TEST);
      gl.glDepthFunc(GL_LEQUAL);

      gl.glDrawArrays(GL_TRIANGLES, 0, 18);
   }

   public void init(GLAutoDrawable drawable)
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      renderingProgram = Utils.createShaderProgram("shaders/vertShader_15.glsl",
                                                   "shaders/fragShader_15.glsl");
      setupVertices();

      aspect = (float) myCanvas.getWidth() / (float) myCanvas.getHeight();
      pMat.identity().setPerspective((float) Math.toRadians(60.0f), aspect, 0.1f, 1000.0f);

      cameraX = 0.0f; cameraY = 0.0f; cameraZ = 4.0f;
      pyrLocX = 0.0f; pyrLocY = 0.0f; pyrLocZ = 0.0f;

      brickTexture = Utils.loadTextureAWT("textures/brick1.jpg");
   }

   private void setupVertices()
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      final float[] pyramidPositions =
      {
        -1.0f, -1.0f,  1.0f,  1.0f, -1.0f,  1.0f,  0.0f,  1.0f,  0.0f, //front
         1.0f, -1.0f,  1.0f,  1.0f, -1.0f, -1.0f,  0.0f,  1.0f,  0.0f, //right
         1.0f, -1.0f, -1.0f, -1.0f, -1.0f, -1.0f,  0.0f,  1.0f,  0.0f, //back
        -1.0f, -1.0f, -1.0f, -1.0f, -1.0f,  1.0f,  0.0f,  1.0f,  0.0f, //left
        -1.0f, -1.0f, -1.0f,  1.0f, -1.0f,  1.0f, -1.0f, -1.0f,  1.0f, //LF
         1.0f, -1.0f,  1.0f, -1.0f, -1.0f, -1.0f,  1.0f, -1.0f, -1.0f  //RR
      };

      final float[] pyrTextureCoordinates =
      {
         0.0f, 0.0f, 1.0f, 0.0f, 0.5f, 1.0f,
         0.0f, 0.0f, 1.0f, 0.0f, 0.5f, 1.0f,
         0.0f, 0.0f, 1.0f, 0.0f, 0.5f, 1.0f,
         0.0f, 0.0f, 1.0f, 0.0f, 0.5f, 1.0f,
         0.0f, 0.0f, 1.0f, 1.0f, 0.0f, 1.0f,
         1.0f, 1.0f, 0.0f, 0.0f, 1.0f, 0.0f
      };

      gl.glGenVertexArrays(vao.length, vao, 0);
      gl.glBindVertexArray(vao[0]);
      gl.glGenBuffers(vbo.length, vbo, 0);

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
      FloatBuffer pyrBuf = Buffers.newDirectFloatBuffer(pyramidPositions);
      gl.glBufferData(GL_ARRAY_BUFFER, pyrBuf.limit()*4, pyrBuf, GL_STATIC_DRAW);

      gl.glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
      FloatBuffer texBuf = Buffers.newDirectFloatBuffer(pyrTextureCoordinates);
      gl.glBufferData(GL_ARRAY_BUFFER, texBuf.limit()*4, texBuf, GL_STATIC_DRAW);
   }

   public void reshape(GLAutoDrawable drawable, int x, int y, int w, int h)
   {
      aspect = (float) myCanvas.getWidth() / (float) myCanvas.getHeight();
      pMat.identity().setPerspective((float)Math.toRadians(60.0f), aspect, 0.1f, 1000.0f);
   }

   public void dispose(GLAutoDrawable drawable){}

   public static void main(String[] args)
   {
      new Code_16();
   }
}
