/*
Computer Graphics Programming in OpenGL with Java, 2nd Edition, V. Scott Gordon, John Clevenger
Pages 11-12.
*/

import static com.jogamp.opengl.GL4.*;
import com.jogamp.opengl.*;
import com.jogamp.opengl.awt.GLJPanel;

import javax.swing.*;

public class Code_2 extends JFrame implements GLEventListener
{
   private final GLJPanel myCanvas;
   private final int[] vao = new int[1];
   private int renderingProgram;

   public Code_2()
   {
      this.setTitle("Chapter 2 - program 2 (Code_2)");
      this.setSize(600, 400);
      myCanvas = new GLJPanel();
      myCanvas.addGLEventListener(this);
      this.add(myCanvas);
      this.setLocationRelativeTo(null);
      this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      this.setVisible(true);
   }


   public void init(GLAutoDrawable drawable)
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      gl.glGenVertexArrays(vao.length, vao, 0);
      gl.glBindVertexArray(vao[0]);

      final String vshaderSource[] =
      {
         "#version 430    \n",
         "void main(void) \n",
         "{ gl_Position = vec4(0.0, 0.0, 0.0, 1.0); } \n"
      };

      final String fshaderSource[] =
      {
         "#version 430    \n",
         "out vec4 color; \n",
         "void main(void) \n",
         "{ color = vec4(0.0, 0.0, 1.0, 1.0); } \n"
      };

      // Compile vertex shader (no error checking!).
      final int vShader = gl.glCreateShader(GL_VERTEX_SHADER);
      gl.glShaderSource(vShader, 3, vshaderSource, null, 0); // 3 is the count of lines of source code
      gl.glCompileShader(vShader);

      // Compile fragment shader (no error checking!).
      final int fShader = gl.glCreateShader(GL_FRAGMENT_SHADER);
      gl.glShaderSource(fShader, 4, fshaderSource, null, 0); // 4 is the count of lines of source code
      gl.glCompileShader(fShader);

      // Link shaders (no error checking!).
      renderingProgram = gl.glCreateProgram();
      gl.glAttachShader(renderingProgram, vShader);
      gl.glAttachShader(renderingProgram, fShader);
      gl.glLinkProgram(renderingProgram);

      gl.glDeleteShader(vShader);
      gl.glDeleteShader(fShader);
   }


   public void display(GLAutoDrawable drawable)
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      gl.glUseProgram(renderingProgram);
      gl.glPointSize(5.0f);
      gl.glDrawArrays(GL_POINTS, 0, 1);
   }

   public void reshape(GLAutoDrawable drawable, int x, int y, int w, int h){}
   public void dispose(GLAutoDrawable drawable){}

   public static void main(String[] args)
   {
      new Code_2();
   }
}
