/*
Computer Graphics Programming in OpenGL with Java, 2nd Edition, V. Scott Gordon, John Clevenger
Pages 26-27.
*/

import static com.jogamp.opengl.GL4.*;
import com.jogamp.opengl.*;
import com.jogamp.opengl.awt.GLJPanel;

import javax.swing.*;
import java.io.File;
import java.io.IOException;
import java.util.Scanner;
import java.util.Vector;

public class Code_4 extends JFrame implements GLEventListener
{
   private final GLJPanel myCanvas;
   private final int[] vao = new int[1];
   private int renderingProgram;

   public Code_4()
   {
      this.setTitle("Chapter 2 - program 4 (Code_4)");
      this.setSize(600, 400);
      myCanvas = new GLJPanel();
      myCanvas.addGLEventListener(this);
      this.add(myCanvas);
      this.setLocationRelativeTo(null);
      this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      this.setVisible(true);
   }


   public void init(GLAutoDrawable drawable)
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      gl.glGenVertexArrays(vao.length, vao, 0);
      gl.glBindVertexArray(vao[0]);

      final String vshaderSource[] = readShaderSource("shaders/vertShader_4.glsl");
      final String fshaderSource[] = readShaderSource("shaders/fragShader_4.glsl");

      // Compile vertex shader (no error checking!).
      final int vShader = gl.glCreateShader(GL_VERTEX_SHADER);
      gl.glShaderSource(vShader, vshaderSource.length, vshaderSource, null, 0);
      gl.glCompileShader(vShader);

      // Compile fragment shader (no error checking!).
      final int fShader = gl.glCreateShader(GL_FRAGMENT_SHADER);
      gl.glShaderSource(fShader, fshaderSource.length, fshaderSource, null, 0);
      gl.glCompileShader(fShader);

      // Link shaders (no error checking!).
      renderingProgram = gl.glCreateProgram();
      gl.glAttachShader(renderingProgram, vShader);
      gl.glAttachShader(renderingProgram, fShader);
      gl.glLinkProgram(renderingProgram);

      gl.glDeleteShader(vShader);
      gl.glDeleteShader(fShader);
   }


   public void display(GLAutoDrawable drawable)
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      gl.glUseProgram(renderingProgram);
      gl.glPointSize(5.0f);
      gl.glDrawArrays(GL_POINTS, 0, 1);
   }

   public void reshape(GLAutoDrawable drawable, int x, int y, int w, int h){}
   public void dispose(GLAutoDrawable drawable){}


   private String[] readShaderSource(String filename)
   {
      final Vector<String> lines = new Vector<String>();
      String[] program;
      try
      {
         final Scanner sc = new Scanner(new File(filename));
         while (sc.hasNext())
         {
            lines.addElement(sc.nextLine());
         }
         program = new String[lines.size()];
         for (int i = 0; i < lines.size(); ++i)
         {
            program[i] = lines.elementAt(i) + "\n";
         }
      }
      catch (IOException e)
      {
         System.err.println("IOException reading file: " + e);
         return null;
      }
      return program;
   }

   public static void main(String[] args)
   {
      new Code_4();
   }
}
