/*
Computer Graphics Programming in OpenGL with Java, 2nd Edition, V. Scott Gordon, John Clevenger
Pages 30-31.
*/

import static com.jogamp.opengl.GL4.*;
import com.jogamp.opengl.*;
import com.jogamp.opengl.awt.GLJPanel;
import com.jogamp.opengl.util.Animator;

import javax.swing.*;

public class Code_6 extends JFrame implements GLEventListener
{
   private final GLJPanel myCanvas;
   private final int[] vao = new int[1];
   private int renderingProgram;

   private float x = 0.0f;
   private float inc = 0.01f;

   public Code_6()
   {
      this.setTitle("Chapter 2 - program 6 (Code_6)");
      this.setSize(400, 200);
      myCanvas = new GLJPanel();
      myCanvas.addGLEventListener(this);
      this.add(myCanvas);
      this.setLocationRelativeTo(null);
      this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      this.setVisible(true);

      final Animator animator = new Animator(myCanvas);
      animator.start();
   }


   public void init(GLAutoDrawable drawable)
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      gl.glGenVertexArrays(vao.length, vao, 0);
      gl.glBindVertexArray(vao[0]);

      renderingProgram = Utils.createShaderProgram("shaders/vertShader_6.glsl",
                                                   "shaders/fragShader_4.glsl");
   }


   public void display(GLAutoDrawable drawable)
   {
      final GL4 gl = (GL4) GLContext.getCurrentGL();

      gl.glClear(GL_COLOR_BUFFER_BIT);
      gl.glClear(GL_DEPTH_BUFFER_BIT);
      gl.glUseProgram(renderingProgram);

      x += inc;
      if (x >  1.0f) inc = -0.01f;
      if (x < -1.0f) inc =  0.01f;
      final int offsetLoc = gl.glGetUniformLocation(renderingProgram, "inc");
      gl.glProgramUniform1f(renderingProgram, offsetLoc, x);

      gl.glDrawArrays(GL_TRIANGLES, 0, 3);
   }

   public void reshape(GLAutoDrawable drawable, int x, int y, int w, int h){}
   public void dispose(GLAutoDrawable drawable){}


   public static void main(String[] args)
   {
      new Code_6();
   }
}
