#version 430

layout (location=0) in vec3 position;

uniform mat4 v_matrix;
uniform mat4 proj_matrix;
uniform float tf;

out vec4 varyingColor;

mat4 buildRotateX(float rad);
mat4 buildRotateY(float rad);
mat4 buildRotateZ(float rad);
mat4 buildTranslate(float x, float y, float z);

void main(void)
{
   float x = gl_InstanceID + tf;

   mat4 localRotX = buildRotateX(1.75 * x);
   mat4 localRotY = buildRotateY(1.75 * x);
   mat4 localRotZ = buildRotateZ(1.75 * x);

  //float dx = sin(.35 * x) * 8.0;            // when 24 instances
  //float dy = cos(.52 * x) * 8.0;
  //float dz = sin(.70 * x) * 8.0;

   float dx = sin(203.0 * x/8000.0) * 403.0;  // when 100,000 instances
   float dy = cos(301.0 * x/4001.0) * 401.0;
   float dz = sin(400.0 * x/6003.0) * 405.0;
   mat4 localTrans = buildTranslate(dx, dy, dz);

   mat4 model_matrix = localTrans * localRotX * localRotY * localRotZ;
   mat4 mv_matrix = v_matrix * model_matrix;
   gl_Position = proj_matrix * mv_matrix * vec4(position,1.0);
   varyingColor = vec4(position,1.0) * 0.5 + vec4(0.5, 0.5, 0.5, 0.5);
}

mat4 buildTranslate(float dx, float dy, float dz)
{	
   mat4 trans = mat4(	
                 1.0, 0.0, 0.0, 0.0,
                 0.0, 1.0, 0.0, 0.0,
                 0.0, 0.0, 1.0, 0.0,
                 dx,  dy,  dz,  1.0 );
   return trans;
}

mat4 buildRotateX(float rad)
{	
   mat4 xrot = mat4(	
                 1.0, 0.0,       0.0,      0.0,
                 0.0, cos(rad), -sin(rad), 0.0,
                 0.0, sin(rad),  cos(rad), 0.0,
                 0.0, 0.0,       0.0,      1.0 );
   return xrot;
}

mat4 buildRotateY(float rad)
{	
   mat4 yrot = mat4(	
                 cos(rad),  0.0,  sin(rad), 0.0,
                 0.0,       1.0,  0.0,      0.0,
                -sin(rad),  0.0,  cos(rad), 0.0,
                 0.0,       0.0,  0.0,      1.0 );
   return yrot;
}

mat4 buildRotateZ(float rad)
{	
   mat4 zrot = mat4(	
                 cos(rad), sin(rad), 0.0, 0.0,
                -sin(rad), cos(rad), 0.0, 0.0,
                 0.0,      0.0,      1.0, 0.0,
                 0.0,      0.0,      0.0, 1.0 );
   return zrot;
}
